@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2004  The DOSBox Team
+ *  Copyright (C) 2002-2006  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -9,13 +9,16 @@
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *  GNU Library General Public License for more details.
+ *  GNU General Public License for more details.
  *
  *  You should have received a copy of the GNU General Public License
  *  along with this program; if not, write to the Free Software
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
+/* $Id: joystick.h,v 1.8 2006/02/09 11:47:48 qbix79 Exp $ */
+#ifndef DOSBOX_JOYSTICK_H
+#define DOSBOX_JOYSTICK_H
 void JOYSTICK_Enable(Bitu which,bool enabled);
 
 void JOYSTICK_Button(Bitu which,Bitu num,bool pressed);
@@ -31,3 +34,14 @@ bool JOYSTICK_GetButton(Bitu which, Bitu num);
 float JOYSTICK_GetMove_X(Bitu which);
 
 float JOYSTICK_GetMove_Y(Bitu which);
+
+enum JoystickType {
+	JOY_NONE,
+	JOY_2AXIS,
+	JOY_4AXIS,
+	JOY_FCS,
+	JOY_CH
+};
+
+extern JoystickType joytype;
+#endif
