@@ -110,8 +110,9 @@ static OpCode OpCodeTable[1024]={
 /* 0x98 - 0x9f */
 {D_CBW		,0			,0		,0		},{D_CWD	,0			,0		,0		},
 {L_Ifw		,O_CALLFw	,0		,0		},{D_WAIT	,0			,0		,0		},
-{L_FLG		,0			,S_PUSHw,0		},{L_POPw	,0			,S_FLGw	,0		},
-{L_REGb		,0			,S_FLGb	,REGI_AH},{L_FLG	,0			,S_REGb	,REGI_AH},
+{D_PUSHF	,0			,0		,0		},{D_POPF	,0			,0		,0		},
+{D_SAHF		,0			,0		,0		},{D_LAHF	,0			,0		,0		},
+
 
 /* 0xa0 - 0xa7 */
 {L_OP		,O_ALOP		,0		,0		},{L_OP		,O_AXOP		,0		,0		},
@@ -170,7 +171,7 @@ static OpCode OpCodeTable[1024]={
 {L_REGw		,O_OUTb		,0		,REGI_DX},{L_REGw	,O_OUTw		,0		,REGI_DX},
 
 /* 0xf0 - 0xf7 */
-{0			,0			,0		,0		},{0		,0			,0		,0		},
+{D_LOCK		,0			,0		,0		},{D_ICEBP	,0			,0		,0		},
 {L_PREREPNE	,0			,0		,0		},{L_PREREP	,0			,0		,0		},
 {D_HLT		,0			,0		,0		},{D_CMC	,0			,0		,0		},
 {L_MODRM	,8			,0		,M_GRP	},{L_MODRM	,9			,0		,M_GRP	},
@@ -182,7 +183,7 @@ static OpCode OpCodeTable[1024]={
 
 /* 0x100 - 0x107 */
 {L_MODRM	,O_GRP6w	,S_Ew	,M_Ew	},{L_MODRM	,O_GRP7w	,S_Ew	,M_Ew	},
-{L_MODRM	,O_LAR		,S_Gw	,M_Ew	},{L_MODRM	,O_LSL		,S_Gw	,M_Ew	},
+{L_MODRM	,O_LAR		,S_Gw	,M_EwGw	},{L_MODRM	,O_LSL		,S_Gw	,M_EwGw	},
 {0			,0			,0		,0		},{0		,0			,0		,0		},
 {D_CLTS		,0			,0		,0		},{0		,0			,0		,0		},
 /* 0x108 - 0x10f */
@@ -315,7 +316,7 @@ static OpCode OpCodeTable[1024]={
 {L_MODRM	,0			,S_Gw	,M_Ebx	},{L_MODRM	,0			,S_Gw	,M_Ewx	},
 
 /* 0x1c0 - 0x1cc */
-{0			,0			,0		,0		},{0		,0			,0		,0		},
+{L_MODRM		,t_ADDb			,S_EbGb		,M_GbEb		},{L_MODRM		,t_ADDw		,S_EwGw		,M_GwEw	},
 {0			,0			,0		,0		},{0		,0			,0		,0		},
 {0			,0			,0		,0		},{0		,0			,0		,0		},
 {0			,0			,0		,0		},{0		,0			,0		,0		},
@@ -466,8 +467,8 @@ static OpCode OpCodeTable[1024]={
 /* 0x298 - 0x29f */
 {D_CWDE		,0			,0		,0		},{D_CDQ	,0			,0		,0		},
 {L_Ifd		,O_CALLFd	,0		,0		},{D_WAIT	,0			,0		,0		},
-{L_FLG		,0			,S_PUSHd,0		},{L_POPd	,0			,S_FLGd	,0		},
-{L_REGb		,0			,S_FLGb	,REGI_AH},{L_FLG	,0			,S_REGb	,REGI_AH},
+{D_PUSHF	,0			,0		,true	},{D_POPF	,0			,0		,true	},
+{D_SAHF		,0			,0		,0		},{D_LAHF	,0			,0		,0		},
 
 /* 0x2a0 - 0x2a7 */
 {L_OP		,O_ALOP		,0		,0		},{L_OP		,O_EAXOP	,0		,0		},
@@ -525,7 +526,7 @@ static OpCode OpCodeTable[1024]={
 {L_REGw		,O_OUTb		,0		,REGI_DX},{L_REGw	,O_OUTd		,0		,REGI_DX},
 
 /* 0x2f0 - 0x2f7 */
-{0			,0			,0		,0		},{0		,0			,0		,0		},
+{D_LOCK		,0			,0		,0		},{D_ICEBP	,0			,0		,0		},
 {L_PREREPNE	,0			,0		,0		},{L_PREREP	,0			,0		,0		},
 {0			,0			,0		,0		},{D_CMC	,0			,0		,0		},
 {L_MODRM	,8			,0		,M_GRP	},{L_MODRM	,0xa		,0		,M_GRP	},
@@ -538,7 +539,7 @@ static OpCode OpCodeTable[1024]={
 
 /* 0x300 - 0x307 */
 {L_MODRM	,O_GRP6d	,S_Ew	,M_Ew	},{L_MODRM	,O_GRP7d	,S_Ew	,M_Ew	},
-{L_MODRM	,O_LAR		,S_Gd	,M_Ew	},{L_MODRM	,O_LSL		,S_Gd	,M_Ew	},
+{L_MODRM	,O_LAR		,S_Gd	,M_EdGd	},{L_MODRM	,O_LSL		,S_Gd	,M_EdGd	},
 {0			,0			,0		,0		},{0		,0			,0		,0		},
 {D_CLTS		,0			,0		,0		},{0		,0			,0		,0		},
 /* 0x308 - 0x30f */
@@ -671,7 +672,7 @@ static OpCode OpCodeTable[1024]={
 {L_MODRM	,0			,S_Gd	,M_Ebx	},{L_MODRM	,0			,S_Gd	,M_Ewx	},
 
 /* 0x3c0 - 0x3cc */
-{0			,0			,0		,0		},{0		,0			,0		,0		},
+{L_MODRM		,t_ADDb			,S_EbGb		,M_GbEb		},{L_MODRM		,t_ADDd		,S_EdGd		,M_GdEd	},
 {0			,0			,0		,0		},{0		,0			,0		,0		},
 {0			,0			,0		,0		},{0		,0			,0		,0		},
 {0			,0			,0		,0		},{0		,0			,0		,0		},
