@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2004  The DOSBox Team
+ *  Copyright (C) 2002-2006  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,7 +16,7 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-/* $Id: drives.h,v 1.25 2004/11/03 23:13:55 qbix79 Exp $ */
+/* $Id: drives.h,v 1.31 2006/02/12 13:32:30 qbix79 Exp $ */
 
 #ifndef _DRIVES_H__
 #define _DRIVES_H__
@@ -264,7 +264,7 @@ struct isoDirEntry {
 #pragma pack ()
 #endif
 
-#if defined (WORD_BIGENDIAN)
+#if defined (WORDS_BIGENDIAN)
 #define EXTENT_LOCATION(de)	((de).extentLocationM)
 #define DATA_LENGTH(de)		((de).dataLengthM)
 #else
@@ -274,9 +274,13 @@ struct isoDirEntry {
 
 #define ISO_FRAMESIZE		2048
 #define ISO_DIRECTORY		2
+#define ISO_HIDDEN		1
+#define ISO_MAX_FILENAME_LENGTH 37
 #define ISO_MAXPATHNAME		256
 #define ISO_FIRST_VD		16
 #define IS_DIR(fileFlags)	(fileFlags & ISO_DIRECTORY)
+#define IS_HIDDEN(fileFlags)	(fileFlags & ISO_HIDDEN)
+#define ISO_MAX_HASH_TABLE_SIZE 	100
 
 class isoDrive : public DOS_Drive {
 public:
@@ -300,14 +304,33 @@ public:
 	virtual bool isRemote(void);
 	virtual bool isRemovable(void);
 	bool readSector(Bit8u *buffer, Bit32u sector);
+	virtual char const* GetLabel(void) {return discLabel;};
 private:
 	int  readDirEntry(isoDirEntry *de, Bit8u *data);
 	bool loadImage();
 	bool lookupSingle(isoDirEntry *de, const char *name, Bit32u sectorStart, Bit32u length);
 	bool lookup(isoDirEntry *de, const char *path);
+	
+	int  GetDirIterator(const isoDirEntry* de);
+	bool GetNextDirEntry(const int dirIterator, isoDirEntry* de);
+	void FreeDirIterator(const int dirIterator);
+	bool ReadCachedSector(Bit8u** buffer, const Bit32u sector);
+	
+	struct DirIterator {
+		bool valid;
+		Bit32u currentSector;
+		Bit32u endSector;
+		Bit32u pos;
+	} dirIterators[MAX_OPENDIRS];
+	
+	int nextFreeDirIterator;
+	
+	struct SectorHashEntry {
+		bool valid;
+		Bit32u sector;
+		Bit8u data[ISO_FRAMESIZE];
+	} sectorHashEntries[ISO_MAX_HASH_TABLE_SIZE];
 
-	std::vector<isoDirEntry> searchCache;
-	std::vector<isoDirEntry>::iterator dirIter;
 	isoDirEntry rootEntry;
 	Bit8u mediaid;
 	Bit8u subUnit;
