@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2004  The DOSBox Team
+ *  Copyright (C) 2002-2006  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,6 +16,8 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
+/* $Id: ipx.cpp,v 1.10 2006/02/26 13:46:31 qbix79 Exp $ */
+
 #include "dosbox.h"
 
 #if C_IPX
@@ -26,7 +28,6 @@
 #include "cross.h"
 #include "support.h"
 #include "cpu.h"
-#include "SDL_net.h"
 #include "regs.h"
 #include "inout.h"
 #include "setup.h"
@@ -39,46 +40,33 @@
 #include "timer.h"
 #include "SDL_net.h"
 #include "programs.h"
+#include "pic.h"
 
 #define SOCKTABLESIZE	150 // DOS IPX driver was limited to 150 open sockets
-#define DOS_MEMSEG		0xd000;
-
-Bit32u tcpPort;
-bool isIpxServer;
-bool isIpxConnected;
-IPaddress ipxClientIp;  // IPAddress for client connection to server
-IPaddress ipxServConnIp;  // IPAddress for client connection to server
-TCPsocket ipxTCPClientSocket;
-UDPsocket ipxClientSocket;
-int UDPChannel;  // Channel used by UDP connection
-Bit8u recvBuffer[IPXBUFFERSIZE]; // Incoming packet buffer
-Bitu call_ipx; // Callback of RETF entrypoint
-Bitu call_ipxint; // Callback of INT 7A entrypoint
-Bitu call_ipxesr1; // Callback of ESR init routine
-Bitu call_ipxesr2; // Callback of ESR return routine
-Bit16u dospage;
-static RealPt ipx_callback;
-static RealPt ipx_intcallback;
-static RealPt ipx_esrcallback;
-static RealPt ipx_esrptraddr;
-static RealPt processedECB;
-
-SDLNet_SocketSet clientSocketSet;
-static bool inESR;
 
 struct ipxnetaddr {
 	Uint8 netnum[4];   // Both are big endian
 	Uint8 netnode[6];
 } localIpxAddr;
 
-struct fragmentDescriptor {
-	Bit16u offset;
-	Bit16u segment;
-	Bit16u size;
-};
+Bit32u tcpPort;
+bool isIpxServer;
+bool isIpxConnected;
+IPaddress ipxClientIp;  // IPAddress for client connection to server
+IPaddress ipxServConnIp;  // IPAddress for client connection to server
+UDPsocket ipxClientSocket;
+int UDPChannel;  // Channel used by UDP connection
+Bit8u recvBuffer[IPXBUFFERSIZE]; // Incoming packet buffer
+
+static RealPt ipx_callback;
+
+SDLNet_SocketSet clientSocketSet;
 
 packetBuffer incomingPacket;
 
+static Bit16u socketCount;
+static Bit16u opensockets[SOCKTABLESIZE]; 
+
 static Bit16u swapByte(Bit16u sockNum) {
 	return (((sockNum>> 8)) | (sockNum << 8));
 }
@@ -93,113 +81,156 @@ void PackIP(IPaddress ipAddr, PackedIP *ipPack) {
 	ipPack->port = ipAddr.port;
 }
 
-class ECBClass {
-public:
-	RealPt ECBAddr;
-    
-	ECBClass *prevECB;
-	ECBClass *nextECB;
-
-	Bit16u getSocket(void) {
-		return swapByte(real_readw(RealSeg(ECBAddr), RealOff(ECBAddr) + 0xa));
-	}
-
-	Bit8u getInUseFlag(void) {
-		return real_readb(RealSeg(ECBAddr), RealOff(ECBAddr) + 0x8);
-	}
-
-	void setInUseFlag(Bit8u flagval) {
-		real_writeb(RealSeg(ECBAddr), RealOff(ECBAddr) + 0x8, flagval);
-	}
-
-	void setCompletionFlag(Bit8u flagval) {
-		real_writeb(RealSeg(ECBAddr), RealOff(ECBAddr) + 0x9, flagval);
-	}
-
-	Bit16u getFragCount(void) {
-		return real_readw(RealSeg(ECBAddr), RealOff(ECBAddr) + 34);
-	}
-
-	void getFragDesc(Bit16u descNum, fragmentDescriptor *fragDesc) {
-		Bit16u memoff = RealOff(ECBAddr) + 30 + ((descNum+1) * 6);
-		fragDesc->offset = real_readw(RealSeg(ECBAddr), memoff);
-		memoff+=2;
-		fragDesc->segment = real_readw(RealSeg(ECBAddr), memoff);
-		memoff+=2;
-		fragDesc->size = real_readw(RealSeg(ECBAddr), memoff);
-	}
-
-	RealPt getESRAddr(void) {
-		return RealMake(real_readw(RealSeg(ECBAddr), RealOff(ECBAddr)+6), real_readw(RealSeg(ECBAddr), RealOff(ECBAddr)+4));
-	}
-
-	void NotifyESR(void) {
-		RealPt tmpAddr = getESRAddr();
-		if(tmpAddr == 0) return;
-		if(!inESR) {
-			//LOG_MSG("Calling ESR");
-			processedECB = ECBAddr;
-			inESR = true;
-			//LOG_MSG("Write %x to %x", real_readd(RealSeg(ECBAddr), RealOff(ECBAddr)+4), RealMake(RealSeg(ipx_esrptraddr), RealOff(ipx_esrptraddr)));
-			real_writed(RealSeg(ipx_esrptraddr), RealOff(ipx_esrptraddr),real_readd(RealSeg(ECBAddr), RealOff(ECBAddr)+4));
-			CPU_CALL(false, RealSeg(ipx_esrcallback), RealOff(ipx_esrcallback),0);
-		}
-	}
-
-	void setImmAddress(Bit8u *immAddr) {
-		Bits i;
-		for(i=0;i<6;i++) {
-			real_writeb(RealSeg(ECBAddr), RealOff(ECBAddr)+28, immAddr[i]);
-		}
-	}
-
-};
-
 ECBClass *ECBList;  // Linked list of ECB's
+ECBClass* ESRList;	// ECBs waiting to be ESR notified
 
-ECBClass * CreateECB(RealPt useAddr) {
-	ECBClass *tmpECB;
-	tmpECB = new ECBClass();
+#ifdef IPX_DEBUGMSG 
+Bitu ECBSerialNumber = 0;
+Bitu ECBAmount = 0;
+#endif
+
+
+ECBClass::ECBClass(Bit16u segment, Bit16u offset) {
+	ECBAddr = RealMake(segment, offset);
 	
-	tmpECB->ECBAddr = useAddr;
-	tmpECB->prevECB = NULL;
-	tmpECB->nextECB = NULL;
+#ifdef IPX_DEBUGMSG
+	SerialNumber = ECBSerialNumber;
+	ECBSerialNumber++;
+	ECBAmount++;
+
+	LOG_MSG("ECB: SN%7d created.   Number of ECBs: %3d, ESR %4x:%4x, ECB %4x:%4x",
+		SerialNumber,ECBAmount,
+		real_readw(RealSeg(ECBAddr),
+		RealOff(ECBAddr)+6),
+		real_readw(RealSeg(ECBAddr),
+		RealOff(ECBAddr)+4),segment,offset);
+#endif
+	isInESRList = false;
+	prevECB = NULL;
+	nextECB = NULL;
 	
-	if (ECBList == NULL) {
-		ECBList = tmpECB;
-	} else {
+	if (ECBList == NULL)
+		ECBList = this;
+	else {
 		// Transverse the list until we hit the end
-		ECBClass *useECB;
-		useECB = ECBList;
-		while(useECB->nextECB != NULL) {
+		ECBClass *useECB = ECBList;
+		
+		while(useECB->nextECB != NULL)
 			useECB = useECB->nextECB;
+
+		useECB->nextECB = this;
+		this->prevECB = useECB;
+	}
+}
+
+Bit16u ECBClass::getSocket(void) {
+	return swapByte(real_readw(RealSeg(ECBAddr), RealOff(ECBAddr) + 0xa));
+}
+
+Bit8u ECBClass::getInUseFlag(void) {
+	return real_readb(RealSeg(ECBAddr), RealOff(ECBAddr) + 0x8);
+}
+
+void ECBClass::setInUseFlag(Bit8u flagval) {
+	iuflag = flagval;
+	real_writeb(RealSeg(ECBAddr), RealOff(ECBAddr) + 0x8, flagval);
+}
+
+void ECBClass::setCompletionFlag(Bit8u flagval) {
+	real_writeb(RealSeg(ECBAddr), RealOff(ECBAddr) + 0x9, flagval);
+}
+
+Bit16u ECBClass::getFragCount(void) {
+	return real_readw(RealSeg(ECBAddr), RealOff(ECBAddr) + 34);
+}
+
+void ECBClass::getFragDesc(Bit16u descNum, fragmentDescriptor *fragDesc) {
+	Bit16u memoff = RealOff(ECBAddr) + 30 + ((descNum+1) * 6);
+	fragDesc->offset = real_readw(RealSeg(ECBAddr), memoff);
+	memoff += 2;
+	fragDesc->segment = real_readw(RealSeg(ECBAddr), memoff);
+	memoff += 2;
+	fragDesc->size = real_readw(RealSeg(ECBAddr), memoff);
+}
+
+RealPt ECBClass::getESRAddr(void) {
+	return RealMake(real_readw(RealSeg(ECBAddr),
+		RealOff(ECBAddr)+6),
+		real_readw(RealSeg(ECBAddr),
+		RealOff(ECBAddr)+4));
+}
+
+void ECBClass::NotifyESR(void) {
+	Bit32u ESRval = real_readd(RealSeg(ECBAddr), RealOff(ECBAddr)+4);
+	if(ESRval) {
+#ifdef IPX_DEBUGMSG 
+		LOG_MSG("ECB: SN%7d to be notified.", SerialNumber);
+#endif
+		// take the ECB out of the current list
+		if(prevECB == NULL) {	// was the first in the list
+			ECBList = nextECB;
+			if(ECBList != NULL) ECBList->prevECB = NULL;
+		} else {		// not the first
+			prevECB->nextECB = nextECB;
+			if(nextECB != NULL) nextECB->prevECB = prevECB;
 		}
-		useECB->nextECB = tmpECB;
-		tmpECB->prevECB = useECB;
-	}
 
-	return tmpECB;
+		nextECB = NULL;
+		// put it to the notification queue
+		if(ESRList==NULL) {
+			ESRList = this;
+			prevECB = NULL;
+		} else  {// put to end of ESR list
+			ECBClass* useECB = ESRList;
+		
+			while(useECB->nextECB != NULL)
+				useECB = useECB->nextECB;
+
+			useECB->nextECB = this;
+			prevECB = useECB;
+		}
+		isInESRList = true;
+		//LOG_MSG("ECB put to notification list");
+		PIC_ActivateIRQ(11);
+	}
+	// this one does not want to be notified, delete it right away
+	else delete this;
 }
 
-void DeleteECB(ECBClass * useECB) {
-	if(useECB == NULL) return;
+void ECBClass::setImmAddress(Bit8u *immAddr) {
+	for(Bitu i=0;i<6;i++)
+		real_writeb(RealSeg(ECBAddr), RealOff(ECBAddr)+28+i, immAddr[i]);
+}
 
-	if(useECB->prevECB == NULL) {
-		ECBList = useECB->nextECB;
-		if(ECBList != NULL) ECBList->prevECB = NULL;
+void ECBClass::getImmAddress(Bit8u* immAddr) {
+	for(Bitu i=0;i<6;i++)
+		immAddr[i] = real_readb(RealSeg(ECBAddr), RealOff(ECBAddr)+28+i);
+}
+
+ECBClass::~ECBClass() {
+#ifdef IPX_DEBUGMSG 
+	ECBAmount--;
+	LOG_MSG("ECB: SN%7d destroyed. Remaining ECBs: %3d", SerialNumber,ECBAmount);
+#endif
+
+	if(isInESRList) {
+		// in ESR list, always the first element is deleted.
+		ESRList=nextECB;
 	} else {
-		useECB->prevECB->nextECB = useECB->nextECB;
-		if(useECB->nextECB != NULL) useECB->nextECB->prevECB = useECB->prevECB;
+		if(prevECB == NULL) {	// was the first in the list
+			ECBList = nextECB;
+			if(ECBList != NULL) ECBList->prevECB = NULL;
+		} else {	// not the first
+			prevECB->nextECB = nextECB;
+			if(nextECB != NULL) nextECB->prevECB = prevECB;
+		}
 	}
-	delete useECB;
 }
 
-Bit16u socketCount;
-Bit16u opensockets[SOCKTABLESIZE]; 
+
 
 static bool sockInUse(Bit16u sockNum) {
-	Bit16u i;
-	for(i=0;i<socketCount;i++) {
+	for(Bitu i=0;i<socketCount;i++) {
 		if (opensockets[i] == sockNum) return true;
 	}
 	return false;
@@ -207,7 +238,7 @@ static bool sockInUse(Bit16u sockNum) {
 
 static void OpenSocket(void) {
 	Bit16u sockNum, sockAlloc;
-
+	//LOG_MSG("IPX: DOSINT used.");
 	sockNum = swapByte(reg_dx);
 
 	if(socketCount >= SOCKTABLESIZE) {
@@ -236,11 +267,12 @@ static void OpenSocket(void) {
 
 	reg_al = 0x00; // Success
 	reg_dx = swapByte(sockNum);  // Convert back to big-endian
-
 }
 
 static void CloseSocket(void) {
 	Bit16u sockNum, i;
+	ECBClass* tmpECB = ECBList;
+	ECBClass* tmp2ECB = ECBList;
 
 	sockNum = swapByte(reg_dx);
 	if(!sockInUse(sockNum)) return;
@@ -253,76 +285,149 @@ static void CloseSocket(void) {
 		}
 	}
 	--socketCount;
+	
+	// delete all ECBs of that socket
+	while(tmpECB!=0) {
+		tmp2ECB = tmpECB->nextECB;
+		if(tmpECB->getSocket()==sockNum) {
+			tmpECB->setCompletionFlag(COMP_CANCELLED);
+			tmpECB->setInUseFlag(USEFLAG_AVAILABLE);
+			delete tmpECB;
+		}
+		tmpECB = tmp2ECB;
+	}
 }
 
-static bool IPX_Multiplex(void)
-{
+//static RealPt IPXVERpointer;
+
+static bool IPX_Multiplex(void) {
 	if(reg_ax != 0x7a00) return false;
 	reg_al = 0xff;
 	SegSet16(es,RealSeg(ipx_callback));
 	reg_di = RealOff(ipx_callback);
+	
+	//reg_bx = RealOff(IPXVERpointer);
+	//reg_cx = RealSeg(ipx_callback);
 	return true;
 }
 
+static void sendPacket(ECBClass* sendecb);
+
 static void handleIpxRequest(void) {
 	ECBClass *tmpECB;
 
 	switch (reg_bx) {
 		case 0x0000:  // Open socket
 			OpenSocket();
+#ifdef IPX_DEBUGMSG
 			LOG_MSG("IPX: Open socket %4x", swapByte(reg_dx));
+#endif
 			break;
 		case 0x0001:  // Close socket
+#ifdef IPX_DEBUGMSG
 			LOG_MSG("IPX: Close socket %4x", swapByte(reg_dx));
+#endif
 			CloseSocket();
 			break;
+		case 0x0002:	// get local target
+			// es:si
+			// Currently no support for multiple networks
+
+			for(Bitu i = 0; i < 6; i++) 
+				real_writeb(SegValue(es),reg_di+i,real_readb(SegValue(es),reg_si+i+4));
+
+			reg_cx=1;	// time ticks expected
+			reg_al=0x00;//success
+			break;
+
 		case 0x0003:  // Send packet
 			if(!incomingPacket.connected) {
-				tmpECB = CreateECB(RealMake(SegValue(es), reg_si));
+				tmpECB = new ECBClass(SegValue(es),reg_si);
 				tmpECB->setInUseFlag(USEFLAG_AVAILABLE);
 				tmpECB->setCompletionFlag(COMP_UNDELIVERABLE);
-				DeleteECB(tmpECB);
+				delete tmpECB;	// not notify?
 				reg_al = 0xff; // Failure
 			} else {
-				tmpECB = CreateECB(RealMake(SegValue(es), reg_si));
+				tmpECB = new ECBClass(SegValue(es),reg_si);
 				tmpECB->setInUseFlag(USEFLAG_SENDING);
+				//LOG_MSG("IPX: Sending packet on %4x", tmpECB->getSocket());
 				reg_al = 0x00; // Success
+				sendPacket(tmpECB);
 			}
-			//LOG_MSG("IPX: Sending packet on %4x", tmpECB->getSocket());
+
 			break;
 		case 0x0004:  // Listen for packet
-			tmpECB = CreateECB(RealMake(SegValue(es), reg_si));
-			if(!sockInUse(tmpECB->getSocket())) {
-				reg_al = 0xff;  // Socket is not open
+			tmpECB = new ECBClass(SegValue(es),reg_si);
+			if(!sockInUse(tmpECB->getSocket())) {  // Socket is not open
+				reg_al = 0xff;
 				tmpECB->setInUseFlag(USEFLAG_AVAILABLE);
 				tmpECB->setCompletionFlag(COMP_HARDWAREERROR);
-				DeleteECB(tmpECB);
+				delete tmpECB;
 			} else {
 				reg_al = 0x00;  // Success
 				tmpECB->setInUseFlag(USEFLAG_LISTENING);
-				//LOG_MSG("IPX: Listen for packet on 0x%4x - ESR address %4x:%4x", tmpECB->getSocket(), RealSeg(tmpECB->getESRAddr()), RealOff(tmpECB->getESRAddr()));
+				/*LOG_MSG("IPX: Listen for packet on 0x%4x - ESR address %4x:%4x",
+					tmpECB->getSocket(),
+					RealSeg(tmpECB->getESRAddr()),
+					RealOff(tmpECB->getESRAddr()));*/
 			}
 			
 			
 			break;
+		case 0x0006:	// cancel operation
+			{
+				RealPt ecbaddress = RealMake(SegValue(es),reg_si);
+				ECBClass* tmpECB= ECBList;
+				ECBClass* tmp2ECB;
+				while(tmpECB) {
+					tmp2ECB=tmpECB->nextECB;
+					if(tmpECB->ECBAddr == ecbaddress) {
+						tmpECB->setInUseFlag(USEFLAG_AVAILABLE);
+						tmpECB->setCompletionFlag(COMP_CANCELLED);
+						delete tmpECB;
+						reg_al=0;	// Success
+#ifdef IPX_DEBUGMSG 
+						LOG_MSG("IPX: ECB canceled.");
+#endif
+						return;
+					}
+					tmpECB=tmp2ECB;
+				}
+				reg_al=0xff;	// Fail
+				break;
+			}
 		case 0x0008:  // Get interval marker
 			// ????
 			break;
 		case 0x0009:  // Get internetwork address
 			{
+#ifdef IPX_DEBUGMSG
 				LOG_MSG("IPX: Get internetwork address %2x:%2x:%2x:%2x:%2x:%2x", localIpxAddr.netnode[5], localIpxAddr.netnode[4], localIpxAddr.netnode[3], localIpxAddr.netnode[2], localIpxAddr.netnode[1], localIpxAddr.netnode[0]);
-				Bit8u * addrptr;
-				Bits i;
-
-				addrptr = (Bit8u *)&localIpxAddr;
-				for(i=0;i<10;i++) {
-                    real_writeb(SegValue(es),reg_si+i,addrptr[i]);
+#endif	
+				Bit8u * addrptr = (Bit8u *)&localIpxAddr;
+				for(Bit16u i=0;i<10;i++) {
+					real_writeb(SegValue(es),reg_si+i,addrptr[i]);
 				}
 				break;
 			}
 		case 0x000a:  // Relinquish control
 			// Idle thingy
 			break;
+		case 0x000b:	// Disconnect from Target
+					// We don't even connect
+		break;
+
+		case 0x0010:  // SPX install check
+			{
+				reg_al=0;	// SPX not installed
+				break;
+			}
+		case 0x001a:	// get driver maximum packet size
+			{
+				reg_cx=0;	// retry count
+				reg_ax=IPXBUFFERSIZE;		// max packet size
+				break;
+			}
 		default:
 			LOG_MSG("Unhandled IPX function: %4x", reg_bx);
 			break;
@@ -342,15 +447,6 @@ Bitu IPX_IntHandler(void) {
 	return CBRET_NONE;
 }
 
-static void disconnectServer(bool unexpected) {
-	
-	// There is no Timer remove code, hence this has to be done manually
-	incomingPacket.connected = false;
-
-	if(unexpected) LOG_MSG("IPX: Server disconnected unexpectedly");
-
-}
-
 static void pingAck(IPaddress retAddr) {
 	IPXHeader regHeader;
 	UDPpacket regPacket;
@@ -375,7 +471,6 @@ static void pingAck(IPaddress retAddr) {
 	regPacket.channel = UDPChannel;
 	
 	result = SDLNet_UDP_Send(ipxClientSocket, regPacket.channel, &regPacket);
-
 }
 
 static void pingSend(void) {
@@ -406,7 +501,6 @@ static void pingSend(void) {
 	if(!result) {
 		LOG_MSG("IPX: SDLNet_UDP_Send: %s\n", SDLNet_GetError());
 	}
-
 }
 
 static void receivePacket(Bit8u *buffer, Bit16s bufSize) {
@@ -424,7 +518,7 @@ static void receivePacket(Bit8u *buffer, Bit16s bufSize) {
 	// Check to see if ping packet
 	if(useSocket == 0x2) {
 		// Is this a broadcast?
-		if((tmpHeader->dest.addr.byIP.host == 0xffffffff) && (tmpHeader->dest.addr.byIP.port = 0xffff)) {
+		if((tmpHeader->dest.addr.byIP.host == 0xffffffff) && (tmpHeader->dest.addr.byIP.port == 0xffff)) {
 			// Yes.  We should return the ping back to the sender
 			IPaddress tmpAddr;
 			UnpackIP(tmpHeader->src.addr.byIP, &tmpAddr);
@@ -433,9 +527,9 @@ static void receivePacket(Bit8u *buffer, Bit16s bufSize) {
 		}
 	}
 
-
 	useECB = ECBList;
-	while(useECB != NULL) {
+	while(useECB != NULL)
+	{
 		nextECB = useECB->nextECB;
 		if(useECB->getInUseFlag() == USEFLAG_LISTENING) {
 			if(useECB->getSocket() == useSocket) {
@@ -451,10 +545,10 @@ static void receivePacket(Bit8u *buffer, Bit16s bufSize) {
 							useECB->setCompletionFlag(COMP_SUCCESS);
 							useECB->setImmAddress(&buffer[22]);  // Write in source node
 							hostaddr = *((Bit32u *)&buffer[24]);
-
-							//LOG_MSG("IPX: Received packet of %d bytes from %d.%d.%d.%d (%x CRC)", bufSize, CONVIP(hostaddr), packetCRC(&buffer[30], bufSize-30));
+#ifdef IPX_DEBUGMSG
+							LOG_MSG("IPX: Received packet of %d bytes from %d.%d.%d.%d (%x CRC)", bufSize, CONVIP(hostaddr), packetCRC(&buffer[30], bufSize-30));
+#endif		
 							useECB->NotifyESR();
-							DeleteECB(useECB);
 							return;
 						}
 					}
@@ -462,158 +556,18 @@ static void receivePacket(Bit8u *buffer, Bit16s bufSize) {
 				if(bufoffset < bufSize) {
 					useECB->setCompletionFlag(COMP_MALFORMED);
 					useECB->NotifyESR();
-					DeleteECB(useECB);
 					return;
 				}
 			}
 		}
 		useECB = nextECB;
 	}
-
+#ifdef IPX_DEBUGMSG 
+	LOG_MSG("IPX: RX Packet loss!");
+#endif
 }
 
-static void sendPacketsTCP(void) {
-	ECBClass *useECB;
-	ECBClass *nextECB;
-	char outbuffer[IPXBUFFERSIZE];
-	fragmentDescriptor tmpFrag; 
-	Bit16u i, fragCount,t;
-	Bit16s packetsize;
-	Bit16u *wordptr;
-	Bits result;
-
-	useECB = ECBList;
-	while(useECB != NULL) {
-		nextECB = useECB->nextECB;
-		if(useECB->getInUseFlag() == USEFLAG_SENDING) {
-			useECB->setInUseFlag(USEFLAG_AVAILABLE);
-			packetsize = 0;
-			fragCount = useECB->getFragCount(); 
-			for(i=0;i<fragCount;i++) {
-				useECB->getFragDesc(i,&tmpFrag);
-				if(i==0) {
-					// Fragment containing IPX header
-					// Must put source address into header
-					Bit8u * addrptr;
-					Bits m;
-
-					addrptr = (Bit8u *)&localIpxAddr.netnode;
-					for(m=0;m<6;m++) {
-						real_writeb(tmpFrag.segment,tmpFrag.offset+m+22,addrptr[m]);
-					}
-				}
-				for(t=0;t<tmpFrag.size;t++) {
-					outbuffer[packetsize] = real_readb(tmpFrag.segment, tmpFrag.offset + t);
-					packetsize++;
-					if(packetsize>=IPXBUFFERSIZE) {
-						LOG_MSG("IPX: Packet size to be sent greater than %d bytes.", IPXBUFFERSIZE);
-						useECB->setCompletionFlag(COMP_MALFORMED);
-						useECB->NotifyESR();
-						DeleteECB(useECB);
-						goto nextECB;
-					}
-				}
-			}
-			result = SDLNet_TCP_Send(ipxTCPClientSocket, &packetsize, 2);
-			if(result != 2) {
-				useECB->setCompletionFlag(COMP_UNDELIVERABLE);
-				useECB->NotifyESR();
-				DeleteECB(useECB);
-				disconnectServer(true); 
-				return;
-			}
-			
-			// Add length and source socket to IPX header
-			wordptr = (Bit16u *)&outbuffer[0];
-			// Blank CRC
-            wordptr[0] = 0xffff;
-			// Length
-			wordptr[1] = swapByte(packetsize);
-			// Source socket
-			wordptr[14] = swapByte(useECB->getSocket());
-
-			result = SDLNet_TCP_Send(ipxTCPClientSocket, &outbuffer[0], packetsize);
-			if(result != packetsize) {
-				useECB->setCompletionFlag(COMP_UNDELIVERABLE);
-				useECB->NotifyESR();
-				DeleteECB(useECB);
-				disconnectServer(true);
-				return;
-			}
-			useECB->setInUseFlag(USEFLAG_AVAILABLE);
-			useECB->setCompletionFlag(COMP_SUCCESS);
-			useECB->NotifyESR();
-			DeleteECB(useECB);
-
-		}
-nextECB:
-
-		useECB = nextECB;
-	}
-
-}
-
-
-static void IPX_TCPClientLoop(void) {
-	Bits result;
-
-	// Check for incoming packets
-	SDLNet_CheckSockets(clientSocketSet,0);
-	
-	if(SDLNet_SocketReady(ipxClientSocket)) {
-		if(!incomingPacket.inPacket) {
-			if(!incomingPacket.waitsize) {
-				result = SDLNet_TCP_Recv(ipxTCPClientSocket, &incomingPacket.packetSize, 2);
-				if(result!=2) {
-					if(result>0) {
-						incomingPacket.waitsize = true;
-						goto finishReceive;
-					} else {
-						disconnectServer(true);
-						goto finishReceive;
-					}
-				}
-				incomingPacket.packetRead = 0;
-			} else {
-				Bit8u * nextchar;
-				nextchar = (Bit8u *)&incomingPacket.packetSize;
-				nextchar++;
-				result = SDLNet_TCP_Recv(ipxTCPClientSocket, nextchar, 1);
-				if(result!=1) {
-					if(result>0) {
-						LOG_MSG("IPX: Packet overrun");
-					} else {
-						disconnectServer(true);
-						goto finishReceive;
-					}
-				}
-				incomingPacket.waitsize = false;
-				incomingPacket.packetRead = 0;
-			}
-			incomingPacket.inPacket = true;
-		}
-		result = SDLNet_TCP_Recv(ipxTCPClientSocket, &incomingPacket.buffer[incomingPacket.packetRead], incomingPacket.packetSize);
-		if (result>0) {
-			incomingPacket.packetRead+=result;
-			incomingPacket.packetSize-=result;
-			if(incomingPacket.packetSize<=0) {
-				// IPX packet is complete.  Now interpret IPX header and try to match to listening ECB
-				receivePacket(&incomingPacket.buffer[0], incomingPacket.packetRead);
-				incomingPacket.inPacket = false;
-			}
-		} else {
-			// Clost active socket
-			disconnectServer(true);
-		}
-
-	}
-
-finishReceive:;
-
-
-}
-
-static void IPX_UDPClientLoop(void) {
+static void IPX_ClientLoop(void) {
 	int numrecv;
 	UDPpacket inPacket;
 	inPacket.data = (Uint8 *)recvBuffer;
@@ -622,105 +576,122 @@ static void IPX_UDPClientLoop(void) {
 
 	// Its amazing how much simpler UDP is than TCP
 	numrecv = SDLNet_UDP_Recv(ipxClientSocket, &inPacket);
-	if(numrecv) {
-		receivePacket(inPacket.data, inPacket.len);
-	}
-	
+	if(numrecv) receivePacket(inPacket.data, inPacket.len);
 }
 
-static void sendPackets() {
-	ECBClass *useECB;
-	ECBClass *nextECB;
-	char outbuffer[IPXBUFFERSIZE];
+
+void DisconnectFromServer(bool unexpected) {
+	if(unexpected) LOG_MSG("IPX: Server disconnected unexpectedly");
+	if(incomingPacket.connected) {
+		incomingPacket.connected = false;
+		TIMER_DelTickHandler(&IPX_ClientLoop);
+		SDLNet_UDP_Close(ipxClientSocket);
+	}
+}
+
+static void sendPacket(ECBClass* sendecb) {
+	Bit8u outbuffer[IPXBUFFERSIZE];
 	fragmentDescriptor tmpFrag; 
 	Bit16u i, fragCount,t;
 	Bit16s packetsize;
 	Bit16u *wordptr;
 	Bits result;
 	UDPpacket outPacket;
-
-	useECB = ECBList;
-	while(useECB != NULL) {
-		nextECB = useECB->nextECB;
-		if(useECB->getInUseFlag() == USEFLAG_SENDING) {
-			useECB->setInUseFlag(USEFLAG_AVAILABLE);
-			packetsize = 0;
-			fragCount = useECB->getFragCount(); 
-			for(i=0;i<fragCount;i++) {
-				useECB->getFragDesc(i,&tmpFrag);
-				if(i==0) {
-					// Fragment containing IPX header
-					// Must put source address into header
-					Bit8u * addrptr;
-					Bits m;
-
-					addrptr = (Bit8u *)&localIpxAddr.netnode;
-					for(m=0;m<6;m++) {
-						real_writeb(tmpFrag.segment,tmpFrag.offset+m+22,addrptr[m]);
-					}
-				}
-				for(t=0;t<tmpFrag.size;t++) {
-					outbuffer[packetsize] = real_readb(tmpFrag.segment, tmpFrag.offset + t);
-					packetsize++;
-					if(packetsize>=IPXBUFFERSIZE) {
-						LOG_MSG("IPX: Packet size to be sent greater than %d bytes.", IPXBUFFERSIZE);
-						useECB->setCompletionFlag(COMP_UNDELIVERABLE);
-						useECB->NotifyESR();
-						DeleteECB(useECB);
-						goto nextECB;
-					}
-				}
-			}
+		
+	sendecb->setInUseFlag(USEFLAG_AVAILABLE);
+	packetsize = 0;
+	fragCount = sendecb->getFragCount(); 
+	for(i=0;i<fragCount;i++) {
+		sendecb->getFragDesc(i,&tmpFrag);
+		if(i==0) {
+			// Fragment containing IPX header
+			// Must put source address into header
+			Bit8u * addrptr;
 			
-			// Add length and source socket to IPX header
-			wordptr = (Bit16u *)&outbuffer[0];
-			// Blank CRC
-            wordptr[0] = 0xffff;
-			// Length
-			wordptr[1] = swapByte(packetsize);
-			// Source socket
-			wordptr[14] = swapByte(useECB->getSocket());
-
-			outPacket.channel = UDPChannel;
-			outPacket.data = (Uint8 *)&outbuffer[0];
-			outPacket.len = packetsize;
-			outPacket.maxlen = packetsize;
-			// Since we're using a channel, we won't send the IP address again
-			result = SDLNet_UDP_Send(ipxClientSocket, UDPChannel, &outPacket);
-			if(result == 0) {
-				LOG_MSG("IPX: Could not send packet: %s", SDLNet_GetError());
-				useECB->setCompletionFlag(COMP_UNDELIVERABLE);
-				useECB->NotifyESR();
-				DeleteECB(useECB);
-				disconnectServer(true);
+			addrptr = (Bit8u *)&localIpxAddr.netnum;
+			for(Bit16u m=0;m<4;m++) {
+				real_writeb(tmpFrag.segment,tmpFrag.offset+m+18,addrptr[m]);
+			}
+			addrptr = (Bit8u *)&localIpxAddr.netnode;
+			for(Bit16u m=0;m<6;m++) {
+				real_writeb(tmpFrag.segment,tmpFrag.offset+m+22,addrptr[m]);
+			}
+		}
+		for(t=0;t<tmpFrag.size;t++) {
+			outbuffer[packetsize] = real_readb(tmpFrag.segment, tmpFrag.offset + t);
+			packetsize++;
+			if(packetsize>=IPXBUFFERSIZE) {
+				LOG_MSG("IPX: Packet size to be sent greater than %d bytes.", IPXBUFFERSIZE);
+				sendecb->setCompletionFlag(COMP_UNDELIVERABLE);
+				sendecb->NotifyESR();
 				return;
 			}
-			useECB->setInUseFlag(USEFLAG_AVAILABLE);
-			useECB->setCompletionFlag(COMP_SUCCESS);
-			useECB->NotifyESR();
-			DeleteECB(useECB);
-
 		}
-nextECB:
+	}
+	
+	// Add length and source socket to IPX header
+	wordptr = (Bit16u *)&outbuffer[0];
+	// Blank CRC
+	wordptr[0] = 0xffff;
+	// Length
+	wordptr[1] = swapByte(packetsize);
+	// Source socket
+	wordptr[14] = swapByte(sendecb->getSocket());
 
-		useECB = nextECB;
+	Bit8u immedAddr[6];
+	sendecb->getImmAddress(immedAddr);
+	// filter out broadcasts and local loopbacks
+	// Real implementation uses the ImmedAddr to check wether this is a broadcast
+
+	bool islocalbroadcast=true;
+	bool isloopback=true;
+
+	Bit8u * addrptr;
+			
+	addrptr = (Bit8u *)&localIpxAddr.netnum;
+	for(Bitu m=0;m<4;m++) {
+		if(addrptr[m]!=outbuffer[m+0x6])isloopback=false;
+	}
+	addrptr = (Bit8u *)&localIpxAddr.netnode;
+	for(Bitu m=0;m<6;m++) {
+		if(addrptr[m]!=outbuffer[m+0xa])isloopback=false;
+		if(immedAddr[m]!=0xff) islocalbroadcast=false;
 	}
 
+	if(!isloopback) {
+		outPacket.channel = UDPChannel;
+		outPacket.data = (Uint8 *)&outbuffer[0];
+		outPacket.len = packetsize;
+		outPacket.maxlen = packetsize;
+		// Since we're using a channel, we won't send the IP address again
+		result = SDLNet_UDP_Send(ipxClientSocket, UDPChannel, &outPacket);
+		
+		if(result == 0) {
+			LOG_MSG("IPX: Could not send packet: %s", SDLNet_GetError());
+			sendecb->setCompletionFlag(COMP_HARDWAREERROR);
+			sendecb->NotifyESR();
+			DisconnectFromServer(true);
+			return;
+		} else {
+			sendecb->setCompletionFlag(COMP_SUCCESS);
+#ifdef IPX_DEBUGMSG
+			LOG_MSG("Packet sent: size: %d",packetsize);
+#endif
+		}
+	}
+	if(isloopback||islocalbroadcast) {
+		receivePacket(&outbuffer[0],packetsize);
+#ifdef IPX_DEBUGMSG
+		LOG_MSG("Packet back: loopback:%d, broadcast:%d",isloopback,isbroadcast);
+#endif
+	}
+	sendecb->NotifyESR();
 }
 
-static void IPX_ClientLoop(void) {
-	IPX_UDPClientLoop();
-
-	// Send outgoing packets
-	sendPackets();
-}
-
-
 static bool pingCheck(IPXHeader * outHeader) {
 	char buffer[1024];
 	Bits result;
 	UDPpacket regPacket;
-
 	IPXHeader *regHeader;
 	regPacket.data = (Uint8 *)buffer;
 	regPacket.maxlen = sizeof(buffer);
@@ -733,8 +704,6 @@ static bool pingCheck(IPXHeader * outHeader) {
 		return true;
 	}
 	return false;
-
-
 }
 
 bool ConnectToServer(char *strAddr) {
@@ -743,7 +712,12 @@ bool ConnectToServer(char *strAddr) {
 	IPXHeader regHeader;
 
 	if(!SDLNet_ResolveHost(&ipxServConnIp, strAddr, (Bit16u)tcpPort)) {
-		
+		// Generate the MAC address.  This is made by zeroing out the first two
+		// octets and then using the actual IP address for the last 4 octets.
+		// This idea is from the IPX over IP implementation as specified in RFC 1234:
+		// http://www.faqs.org/rfcs/rfc1234.html
+
+
 		// Select an anonymous UDP port
 		ipxClientSocket = SDLNet_UDP_Open(0);
 		if(ipxClientSocket) {
@@ -802,14 +776,12 @@ bool ConnectToServer(char *strAddr) {
 
 				LOG_MSG("IPX: Connected to server.  IPX address is %d:%d:%d:%d:%d:%d", CONVIPX(localIpxAddr.netnode));
 
-
 				incomingPacket.connected = true;
 				TIMER_AddTickHandler(&IPX_ClientLoop);
 				return true;
 			}
 		} else {
 			LOG_MSG("IPX: Unable to open socket");
-
 		}
 	} else {
 		LOG_MSG("IPX: Unable resolve connection to server");
@@ -817,45 +789,21 @@ bool ConnectToServer(char *strAddr) {
 	return false;
 }
 
-void DisconnectFromServer(void) {
+void IPX_NetworkInit() {
 
-	if(incomingPacket.connected) {
-		incomingPacket.connected = false;
-		TIMER_DelTickHandler(&IPX_ClientLoop);
-		SDLNet_UDP_Close(ipxClientSocket);
-	}
-}
-
-bool IPX_NetworkInit() {
-
-	localIpxAddr.netnum[0] = 0x0; localIpxAddr.netnum[1] = 0x0; localIpxAddr.netnum[2] = 0x0; localIpxAddr.netnum[3] = 0x1;
-
-	/*
-	if(SDLNet_ResolveHost(&ipxClientIp, localhostname, tcpPort)) {
-		LOG_MSG("IPX: Unable to resolve localname: \"%s\".  IPX disabled.", localhostname);
-		return false;
-	} else {
-		LOG_MSG("IPX: Using localname: %s IP is: %d.%d.%d.%d", localhostname, CONVIP(ipxClientIp.host));
-	}
-	*/
-
-	// Generate the MAC address.  This is made by zeroing out the first two octets and then using the actual IP address for
-	// the last 4 octets.  This idea is from the IPX over IP implementation as specified in RFC 1234:
-	// http://www.faqs.org/rfcs/rfc1234.html
+	localIpxAddr.netnum[0] = 0x0;
+	localIpxAddr.netnum[1] = 0x0;
+	localIpxAddr.netnum[2] = 0x0;
+	localIpxAddr.netnum[3] = 0x1;
 	localIpxAddr.netnode[0] = 0x00;
 	localIpxAddr.netnode[1] = 0x00;
-	//localIpxAddr.netnode[5] = (ipxClientIp.host >> 24) & 0xff;
-	//localIpxAddr.netnode[4] = (ipxClientIp.host >> 16) & 0xff;
-	//localIpxAddr.netnode[3] = (ipxClientIp.host >> 8) & 0xff;
-	//localIpxAddr.netnode[2] = (ipxClientIp.host & 0xff);
-	//To be filled in on response from server
 	localIpxAddr.netnode[2] = 0x00;
 	localIpxAddr.netnode[3] = 0x00;
 	localIpxAddr.netnode[4] = 0x00;
 	localIpxAddr.netnode[5] = 0x00;
 
 	socketCount = 0;
-	return true;
+	return;
 }
 
 class IPXNET : public Program {
@@ -972,7 +920,7 @@ public:
 					WriteOut("IPX Tunneling Server not running in this DosBox session.\n");
 				} else {
 					isIpxServer = false;
-					DisconnectFromServer();
+					DisconnectFromServer(false);
 					IPX_StopServer();
 					WriteOut("IPX Tunneling Server stopped.");
 					// Don't know how to stop the timer just yet.
@@ -1013,7 +961,7 @@ public:
 				// TODO: Send a packet to the server notifying of disconnect
 
 				WriteOut("IPX Tunneling Client disconnected from server.\n");
-				DisconnectFromServer();
+				DisconnectFromServer(false);
 				return;
 			}
 
@@ -1061,23 +1009,7 @@ public:
 				}
 				return;
 			}
-
-
-
 		}
-
-		/*
-		WriteOut("IPX Status\n\n");
-		if(!incomingPacket.connected) {
-			WriteOut("IPX tunneling client not presently connected");
-			return;
-		}
-		if(isIpxServer) {
-			WriteOut("This DosBox session is an IPX tunneling server running on port %d", tcpPort);
-		} else {
-			WriteOut("This DosBox session is an IPX tunneling client connected to: %s",SDLNet_ResolveIP(&ipxServConnIp));
-		}
-		*/
 	}
 };
 
@@ -1086,103 +1018,147 @@ static void IPXNET_ProgramStart(Program * * make) {
 }
 
 Bitu IPX_ESRHandler1(void) {
-	CPU_Push32(reg_flags);
-	CPU_Push32(reg_eax);CPU_Push32(reg_ecx);CPU_Push32(reg_edx);CPU_Push32(reg_ebx);
-	CPU_Push32(reg_ebp);CPU_Push32(reg_esi);CPU_Push32(reg_edi);
-	CPU_Push16(SegValue(ds)); CPU_Push16(SegValue(es)); 
-
-	SegSet16(es, RealSeg(processedECB));
-	reg_si = RealOff(processedECB);
-	reg_al = 0xff;
-	//LOG_MSG("ESR Callback 1");
+#ifdef IPX_DEBUGMSG 
+	LOG_MSG("ESRhandler entered" );
+#endif
+	while(ESRList!=NULL) {
+#ifdef IPX_DEBUGMSG 
+		LOG_MSG("ECB: SN%7d notified.", ESRList->SerialNumber);
+#endif
+		
+		// setup registers
+		SegSet16(es, RealSeg(ESRList->ECBAddr));
+		reg_si = RealOff(ESRList->ECBAddr);
+		reg_al = 0xff;
+		CALLBACK_RunRealFar(RealSeg(ESRList->getESRAddr()), RealOff(ESRList->getESRAddr()));
+		delete ESRList;
+	}	// while
 
+	IO_WriteB(0xa0,0x63);	//EOI11
+	IO_WriteB(0x20,0x62);	//EOI2
+	
+#ifdef IPX_DEBUGMSG 
+	LOG_MSG("ESR loop done.");
+#endif
+	
 	return CBRET_NONE;
 }
 
-Bitu IPX_ESRHandler2(void) {
-	SegSet16(es, CPU_Pop16()); SegSet16(ds, CPU_Pop16());
-	reg_edi=CPU_Pop32();reg_esi=CPU_Pop32();reg_ebp=CPU_Pop32();
-	reg_ebx=CPU_Pop32();reg_edx=CPU_Pop32();reg_ecx=CPU_Pop32();reg_eax=CPU_Pop32();
-	reg_flags=CPU_Pop32();
+void VFILE_Remove(const char *name);
 
-	//LOG_MSG("Leaving ESR");
-	inESR = false;
+class IPX: public Module_base {
+private:
+	CALLBACK_HandlerObject callback_ipx;
+	CALLBACK_HandlerObject callback_esr;
+	CALLBACK_HandlerObject callback_ipxint;
+	RealPt old_73_vector;
+	static Bit16u dospage;
+public:
+	IPX(Section* configuration):Module_base(configuration) {
+		Section_prop * section = static_cast<Section_prop *>(configuration);
+		if(!section->Get_bool("ipx")) return;
+		if(!SDLNetInited) {
+			if(SDLNet_Init() == -1){
+				LOG_MSG("SDLNet_Init failed: %s\n", SDLNet_GetError());
+				return;
+			}
+			SDLNetInited = true;
+		}
 
-	return CBRET_NONE;
-}
+		ECBList = NULL;
+		ESRList = NULL;
+		isIpxServer = false;
+		isIpxConnected = false;
+		IPX_NetworkInit();
 
-bool IPX_ESRSetupHook(Bitu callback1, CallBack_Handler handler1, Bitu callback2, CallBack_Handler handler2, RealPt *ptrAddr) {
-	PhysPt phyDospage;
-	phyDospage = PhysMake(dospage,0);
+		DOS_AddMultiplexHandler(IPX_Multiplex);
 
-	// Inital callback routine (should save registers, etc.)
-	phys_writeb(phyDospage+0,(Bit8u)0xFA);	//CLI
-	phys_writeb(phyDospage+1,(Bit8u)0xFE);	//GRP 4
-	phys_writeb(phyDospage+2,(Bit8u)0x38);	//Extra Callback instruction
-	phys_writew(phyDospage+3,callback1);		//The immediate word
-	phys_writeb(phyDospage+5,(Bit8u)0x9a);	//CALL Ap
-	// 0x6, 0x7, 0x8, 0x9 = address of called routine
-	*ptrAddr = RealMake(dospage, 6);
-	phys_writed(phyDospage+6,(Bit32u)0x00000000); // Called address
-	phys_writeb(phyDospage+0xa,(Bit8u)0xFE);	//GRP 4
-	phys_writeb(phyDospage+0xb,(Bit8u)0x38);	//Extra Callback instruction
-	phys_writew(phyDospage+0xc,callback2);	//The immediate word
-	phys_writeb(phyDospage+0xe,(Bit8u)0xFB);	//STI
-	phys_writeb(phyDospage+0xf,(Bit8u)0xCB);	//A RETF Instruction
+		callback_ipx.Install(&IPX_Handler,CB_RETF,"IPX Handler");
+		ipx_callback = callback_ipx.Get_RealPointer();
 
-	CallBack_Handlers[callback1]=handler1;
-	CallBack_Handlers[callback2]=handler2;
-	return true;
+		callback_ipxint.Install(&IPX_IntHandler,CB_IRET,"IPX (int 7a)");
+		callback_ipxint.Set_RealVec(0x7a);
+
+		callback_esr.Allocate(&IPX_ESRHandler1,"IPX_ESR");
+		Bit16u call_ipxesr1 = callback_esr.Get_callback();
+
+		if(!dospage) dospage = DOS_GetMemory(2); // can not be freed yet
+
+		PhysPt phyDospage = PhysMake(dospage,0);
+
+#ifdef IPX_DEBUGMSG
+		LOG_MSG("IPX INT address: %x", ipx_intcallback);
+		LOG_MSG("ESR callback address: %x, HandlerID %d", phyDospage,call_ipxesr1);
+#endif
+		//save registers
+		phys_writeb(phyDospage+0,(Bit8u)0xFA);    // CLI
+		phys_writeb(phyDospage+1,(Bit8u)0x60);    // PUSHA
+		phys_writeb(phyDospage+2,(Bit8u)0x1E);    // PUSH DS
+		phys_writeb(phyDospage+3,(Bit8u)0x06);    // PUSH ES
+		phys_writew(phyDospage+4,(Bit16u)0xA00F); // PUSH FS
+		phys_writew(phyDospage+6,(Bit16u)0xA80F); // PUSH GS
+ 
+		// callback
+		phys_writeb(phyDospage+8,(Bit8u)0xFE);  // GRP 4
+		phys_writeb(phyDospage+9,(Bit8u)0x38);  // Extra Callback instruction
+		phys_writew(phyDospage+10,call_ipxesr1);        // Callback identifier
+ 
+		// register recreation
+		phys_writew(phyDospage+12,(Bit16u)0xA90F); // POP GS
+		phys_writew(phyDospage+14,(Bit16u)0xA10F); // POP FS
+		phys_writeb(phyDospage+16,(Bit8u)0x07);    // POP ES
+		phys_writeb(phyDospage+17,(Bit8u)0x1F);    // POP DS
+		phys_writeb(phyDospage+18,(Bit8u)0x61);    // POPA
+		phys_writeb(phyDospage+19,(Bit8u)0xCF);    // IRET: restores flags, CS, IP
+ 
+		// IPX version 2.12
+		//phys_writeb(phyDospage+27,(Bit8u)0x2);
+		//phys_writeb(phyDospage+28,(Bit8u)0x12);
+		//IPXVERpointer = RealMake(dospage,27);
+
+		RealPt ESRRoutineBase = RealMake(dospage, 0);
+
+		// Interrupt enabling
+		RealSetVec(0x73,ESRRoutineBase,old_73_vector);	// IRQ11
+		IO_WriteB(0xa1,IO_ReadB(0xa1)&(~8));			// enable IRQ11
+
+		PROGRAMS_MakeFile("IPXNET.COM",IPXNET_ProgramStart);
+	}
+
+	~IPX() {
+		Section_prop * section = static_cast<Section_prop *>(m_configuration);
+		if(!section->Get_bool("ipx")) return;
+
+		if(isIpxServer) {
+			isIpxServer = false;
+			IPX_StopServer();
+		}
+		DisconnectFromServer(false);
+
+		DOS_DelMultiplexHandler(IPX_Multiplex);
+		RealSetVec(0x73,old_73_vector);
+		IO_WriteB(0xa1,IO_ReadB(0xa1)|8);	// disable IRQ11
+   
+		PhysPt phyDospage = PhysMake(dospage,0);
+		for(Bitu i = 0;i < 32;i++)
+			phys_writeb(phyDospage+i,(Bit8u)0x00);
+
+		VFILE_Remove("IPXNET.COM");
+	}
+};
+
+static IPX* test;
+
+void IPX_ShutDown(Section* sec) {
+	delete test;    
 }
 
 void IPX_Init(Section* sec) {
-	Section_prop * section=static_cast<Section_prop *>(sec);
-
-	if(!section->Get_bool("ipx")) return;
-
-	if(!SDLNetInited) {
-		if(SDLNet_Init()==-1) {
-			LOG_MSG("SDLNet_Init failed: %s\n", SDLNet_GetError());
-			return;
-		}
-		SDLNetInited = true;
-	}
-
-	ECBList = NULL;
-
-	isIpxServer = false;
-	isIpxConnected = false;
-
-	IPX_NetworkInit();
-
-	inESR = false;
-
-	DOS_AddMultiplexHandler(IPX_Multiplex);
-	call_ipx=CALLBACK_Allocate();
-	CALLBACK_Setup(call_ipx,&IPX_Handler,CB_RETF);
-	ipx_callback=CALLBACK_RealPointer(call_ipx);
-
-	call_ipxint=CALLBACK_Allocate();	
-	CALLBACK_Setup(call_ipxint,&IPX_IntHandler,CB_IRET);
-	ipx_intcallback=CALLBACK_RealPointer(call_ipxint);
-
-	call_ipxesr1=CALLBACK_Allocate();	
-	call_ipxesr2=CALLBACK_Allocate();	
-	//CALLBACK_SetupFarCall(call_ipxesr1, &IPX_ESRHandler1, call_ipxesr2, &IPX_ESRHandler2, &ipx_esrptraddr);
-	dospage = DOS_GetMemory(1);
-	IPX_ESRSetupHook(call_ipxesr1, &IPX_ESRHandler1, call_ipxesr2, &IPX_ESRHandler2, &ipx_esrptraddr);
-	// Allocate 16 bytes of memory from the DOS system area at 0xd000
-	ipx_esrcallback=RealMake(dospage,0);
-	//CALLBACK_RealPointer(call_ipxesr1);
-
-	RealSetVec(0x7a,ipx_intcallback);
-	PROGRAMS_MakeFile("IPXNET.COM",IPXNET_ProgramStart);
-
-	//if(isIpxServer) {
-		// Auto-connect to server
-	//	ConnectToServer("localhost");
-	//}
-
+	test = new IPX(sec);
+	sec->AddDestroyFunction(&IPX_ShutDown,true);
 }
 
+//Initialize static members;
+Bit16u IPX::dospage = 0;
+
 #endif
