@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2004  The DOSBox Team
+ *  Copyright (C) 2002-2006  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -23,13 +23,14 @@
 
 static Bit8u flip=0;
 
-void write_p3d4_vga(Bitu port,Bitu val,Bitu iolen);
-Bitu read_p3d4_vga(Bitu port,Bitu iolen);
-void write_p3d5_vga(Bitu port,Bitu val,Bitu iolen);
-Bitu read_p3d5_vga(Bitu port,Bitu iolen);
+void vga_write_p3d4(Bitu port,Bitu val,Bitu iolen);
+Bitu vga_read_p3d4(Bitu port,Bitu iolen);
+void vga_write_p3d5(Bitu port,Bitu val,Bitu iolen);
+Bitu vga_read_p3d5(Bitu port,Bitu iolen);
 
-static Bitu read_p3da(Bitu port,Bitu iolen) {
+static Bitu vga_read_p3da(Bitu port,Bitu iolen) {
 	vga.internal.attrindex=false;
+	vga.tandy.pcjr_flipflop=false;
 	if (vga.config.retrace) {
 		switch (machine) {
 		case MCH_HERC:
@@ -52,23 +53,26 @@ static Bitu read_p3da(Bitu port,Bitu iolen) {
 static void write_p3c2(Bitu port,Bitu val,Bitu iolen) {
 	vga.misc_output=val;
 	if (val & 0x1) {
-		IO_RegisterWriteHandler(0x3d4,write_p3d4_vga,IO_MB);
-		IO_RegisterReadHandler(0x3d4,read_p3d4_vga,IO_MB);
-		IO_RegisterWriteHandler(0x3d5,write_p3d5_vga,IO_MB);
-		IO_RegisterReadHandler(0x3d5,read_p3d5_vga,IO_MB);
-		IO_RegisterReadHandler(0x3da,read_p3da,IO_MB);
-		
+		IO_RegisterWriteHandler(0x3d4,vga_write_p3d4,IO_MB);
+		IO_RegisterReadHandler(0x3d4,vga_read_p3d4,IO_MB);
+		IO_RegisterReadHandler(0x3da,vga_read_p3da,IO_MB);
+
+		IO_RegisterWriteHandler(0x3d5,vga_write_p3d5,IO_MB);
+		IO_RegisterReadHandler(0x3d5,vga_read_p3d5,IO_MB);
+
 		IO_FreeWriteHandler(0x3b4,IO_MB);
 		IO_FreeReadHandler(0x3b4,IO_MB);
 		IO_FreeWriteHandler(0x3b5,IO_MB);
 		IO_FreeReadHandler(0x3b5,IO_MB);
 		IO_FreeReadHandler(0x3ba,IO_MB);
 	} else {
-		IO_RegisterWriteHandler(0x3b4,write_p3d4_vga,IO_MB);
-		IO_RegisterReadHandler(0x3b4,read_p3d4_vga,IO_MB);
-		IO_RegisterWriteHandler(0x3b5,write_p3d5_vga,IO_MB);
-		IO_RegisterReadHandler(0x3b5,read_p3d5_vga,IO_MB);
-		IO_RegisterReadHandler(0x3ba,read_p3da,IO_MB);
+		IO_RegisterWriteHandler(0x3b4,vga_write_p3d4,IO_MB);
+		IO_RegisterReadHandler(0x3b4,vga_read_p3d4,IO_MB);
+		IO_RegisterReadHandler(0x3ba,vga_read_p3da,IO_MB);
+
+		IO_RegisterWriteHandler(0x3b5,vga_write_p3d5,IO_MB);
+		IO_RegisterReadHandler(0x3b5,vga_read_p3d5,IO_MB);
+
 
 		IO_FreeWriteHandler(0x3d4,IO_MB);
 		IO_FreeReadHandler(0x3d4,IO_MB);
@@ -94,18 +98,36 @@ static Bitu read_p3cc(Bitu port,Bitu iolen) {
 	return vga.misc_output;
 }
 
+/*
 
+Test 13:  Hardware: General Registers
+  Mode 00h, General -- Feat Ctrl: IBM=00h   Current=FFh
+  Mode 00h, General -- Input Status 0: IBM=70h   Current=FFh
+  .. & modes 1,2,3,4,5,6,d,e,10,11,12,13
+
+following read handlers silence the above vgatest errors.
+
+*/
+
+static Bitu read_p3ca(Bitu port,Bitu iolen) {
+	return 0;
+}
+
+static Bitu read_p3c2(Bitu port,Bitu iolen) {
+	return 0x70; 
+}
 
 void VGA_SetupMisc(void) {
 	if (machine==MCH_VGA) {
+		IO_RegisterReadHandler(0x3ca,read_p3ca,IO_MB);
+		IO_RegisterReadHandler(0x3c2,read_p3c2,IO_MB);
 		IO_RegisterWriteHandler(0x3c2,write_p3c2,IO_MB);
 		IO_RegisterReadHandler(0x3cc,read_p3cc,IO_MB);
-	} else if (machine==MCH_CGA || machine==MCH_TANDY) {
-		IO_RegisterReadHandler(0x3da,read_p3da,IO_MB);
+	} else if (machine==MCH_CGA || IS_TANDY_ARCH) {
+		IO_RegisterReadHandler(0x3da,vga_read_p3da,IO_MB);
 	} else if (machine==MCH_HERC) {
-		IO_RegisterReadHandler(0x3ba,read_p3da,IO_MB);
+		IO_RegisterReadHandler(0x3ba,vga_read_p3da,IO_MB);
 	}
-
 }
 
 
