@@ -1,5 +1,5 @@
 /*
- *  Copyright (C) 2002-2004  The DOSBox Team
+ *  Copyright (C) 2002-2006  The DOSBox Team
  *
  *  This program is free software; you can redistribute it and/or modify
  *  it under the terms of the GNU General Public License as published by
@@ -16,40 +16,115 @@
  *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-/* $Id: shell.cpp,v 1.53 2004/11/13 12:19:43 qbix79 Exp $ */
+/* $Id: shell.cpp,v 1.72 2006/02/26 15:58:49 qbix79 Exp $ */
 
 #include <stdlib.h>
 #include <stdarg.h>
 #include <string.h>
+#include "dosbox.h"
+#include "regs.h"
 #include "setup.h"
 #include "shell.h"
+#include "callback.h"
+#include "support.h"
 
 Bitu call_shellstop;
+/* Larger scope so shell_del autoexec can use it to
+ * remove things from the environment */
+Program * first_shell = 0; 
 
 static Bitu shellstop_handler(void) {
 	return CBRET_STOP;
 }
 
 static void SHELL_ProgramStart(Program * * make) {
-	*make=new DOS_Shell;
+	*make = new DOS_Shell;
 }
 
 #define AUTOEXEC_SIZE 4096
-static char autoexec_data[AUTOEXEC_SIZE]={0};
+static char autoexec_data[AUTOEXEC_SIZE] = { 0 };
+static std::list<std::string> autoexec_strings;
+typedef std::list<std::string>::iterator auto_it;
 
-void SHELL_AddAutoexec(char * line,...) {
-	char buf[2048];
+void VFILE_Remove(const char *name);
+
+void AutoexecObject::Install(char* line,...) {
+	if(GCC_UNLIKELY(installed)) E_Exit("autoexec: allready created %s",buf);
+	installed = true;
 	va_list msg;
 	
 	va_start(msg,line);
 	vsprintf(buf,line,msg);
 	va_end(msg);
+	autoexec_strings.push_back(std::string(buf));
 
-	size_t auto_len=strlen(autoexec_data);
-	if ((auto_len+strlen(line)+3)>AUTOEXEC_SIZE) {
-		E_Exit("SYSTEM:Autoexec.bat file overflow");
+	this->CreateAutoexec();
+
+	//autoexec.bat is normally created AUTOEXEC_Init.
+	//But if we are allready running (first_shell)
+	//we have to update the envirionment to display changes
+
+	if(first_shell)	{
+		char buf2[256]; strcpy(buf2,buf);//used in shell.h
+		if((strncasecmp(buf2,"set ",4) == 0) && (strlen(buf2) > 4)){
+			char* after_set = buf2 + 4;//move to variable that is being set
+			char* test = strpbrk(after_set,"=");
+			if(!test) {first_shell->SetEnv(after_set,"");return;}
+			*test++ = 0;
+			//If the shell is running/exists update the environment
+			first_shell->SetEnv(after_set,test);
+		}
 	}
-	sprintf((autoexec_data+auto_len),"%s\r\n",buf);
+}
+
+void AutoexecObject::InstallBefore(char* line,...) {
+	if(GCC_UNLIKELY(installed)) E_Exit("autoexec: allready created %s",buf);
+	installed = true;
+	va_list msg;
+	
+	va_start(msg,line);
+	vsprintf(buf,line,msg);
+	va_end(msg);
+	autoexec_strings.push_front(std::string(buf));
+	this->CreateAutoexec();
+}
+
+void AutoexecObject::CreateAutoexec(void) {
+	/* Remove old autoexec.bat if the shell exists */
+	if(first_shell)	VFILE_Remove("AUTOEXEC.BAT");
+
+	//Create a new autoexec.bat
+	autoexec_data[0] = 0;
+	size_t auto_len;
+	for(auto_it it=  autoexec_strings.begin(); it != autoexec_strings.end(); it++) {
+		auto_len = strlen(autoexec_data);
+		if ((auto_len+strlen((*it).c_str())+3)>AUTOEXEC_SIZE) {
+			E_Exit("SYSTEM:Autoexec.bat file overflow");
+		}
+		sprintf((autoexec_data+auto_len),"%s\r\n",(*it).c_str());
+	}
+	if(first_shell) VFILE_Register("AUTOEXEC.BAT",(Bit8u *)autoexec_data,strlen(autoexec_data));
+}
+
+AutoexecObject::~AutoexecObject(){
+	if(!installed) return;
+
+	// Remove the line from the autoexecbuffer and update environment
+	for(auto_it it = autoexec_strings.begin(); it != autoexec_strings.end(); ) {
+		if((*it) == buf) {
+			it = autoexec_strings.erase(it);
+			// If it's a environment variable remove it from there as well
+			if((strncasecmp(buf,"set ",4) == 0) && (strlen(buf) > 4)){
+				char* after_set = buf + 4;//move to variable that is being set
+				char* test = strpbrk(after_set,"=");
+				if(!test) continue;
+				*test = 0;
+				//If the shell is running/exists update the environment
+				if(first_shell) first_shell->SetEnv(after_set,"");
+			}
+		} else it++;
+	}
+	this->CreateAutoexec();
 }
 
 DOS_Shell::DOS_Shell():Program(){
@@ -61,18 +136,24 @@ DOS_Shell::DOS_Shell():Program(){
 	completion_start = NULL;
 }
 
-
-
-
 Bitu DOS_Shell::GetRedirection(char *s, char **ifn, char **ofn,bool * append) {
 
 	char * lr=s;
 	char * lw=s;
 	char ch;
 	Bitu num=0;
+	bool quote = false;
 
 	while ( (ch=*lr++) ) {
+		if(quote && ch != '"') { /* don't parse redirection within quotes. Not perfect yet. Escaped quotes will mess the count up */
+			*lw++ = ch;
+			continue;
+		}
+
 		switch (ch) {
+		case '"':
+			quote = !quote;
+			break;
 		case '>':
 			*append=((*lr)=='>');
 			if (*append) lr++;
@@ -80,6 +161,8 @@ Bitu DOS_Shell::GetRedirection(char *s, char **ifn, char **ofn,bool * append) {
 			if (*ofn) free(*ofn);
 			*ofn=lr;
 			while (*lr && *lr!=' ') lr++;
+			//if it ends on a : => remove it.
+			if((*ofn != lr) && (lr[-1] == ':')) lr[-1] = 0;
 			if(*lr && *(lr+1)) 
 				*lr++=0; 
 			else 
@@ -91,6 +174,7 @@ Bitu DOS_Shell::GetRedirection(char *s, char **ifn, char **ofn,bool * append) {
 			lr=ltrim(lr);
 			*ifn=lr;
 			while (*lr && *lr!=' ') lr++;
+			if((*ifn != lr) && (lr[-1] == ':')) lr[-1] = 0;
 			if(*lr && *(lr+1)) 
 				*lr++=0; 
 			else 
@@ -110,49 +194,67 @@ Bitu DOS_Shell::GetRedirection(char *s, char **ifn, char **ofn,bool * append) {
 void DOS_Shell::ParseLine(char * line) {
 	LOG(LOG_EXEC,LOG_ERROR)("Parsing command line: %s",line);
 	/* Check for a leading @ */
- 	if (line[0]=='@') line[0]=' ';
-	line=trim(line);
+ 	if (line[0] == '@') line[0] = ' ';
+	line = trim(line);
 
 	/* Do redirection and pipe checks */
 	
-	char * in=0;
-	char * out=0;
+	char * in  = 0;
+	char * out = 0;
 
-	Bit16u old_in,old_out;
 	Bit16u dummy,dummy2;
-	Bitu num=0;		/* Number of commands in this line */
+	Bit32u bigdummy = 0;
+	Bitu num = 0;		/* Number of commands in this line */
 	bool append;
-
+	bool normalstdin  = false;	/* wether stdin/out are open on start. */
+	bool normalstdout = false;	/* Bug: Assumed is they are "con"      */
+	
 	num = GetRedirection(line,&in, &out,&append);
 	if (num>1) LOG_MSG("SHELL:Multiple command on 1 line not supported");
-//	if (in || num>1) DOS_DuplicateEntry(0,&old_in);
-
+	if (in || out) {
+			normalstdin  = (psp->GetFileHandle(0) != 0xff); 
+			normalstdout = (psp->GetFileHandle(1) != 0xff); 
+	}
 	if (in) {
 		if(DOS_OpenFile(in,0,&dummy)) { //Test if file exists
 			DOS_CloseFile(dummy);
 			LOG_MSG("SHELL:Redirect input from %s",in);
-			DOS_CloseFile(0); //Close stdin
-			DOS_OpenFile(in,0,&dummy);//Open new stdin
+			if(normalstdin) DOS_CloseFile(0); //Close stdin
+			DOS_OpenFile(in,0,&dummy); //Open new stdin
 		}
 	}
 	if (out){
 		LOG_MSG("SHELL:Redirect output to %s",out);
-		DOS_CloseFile(1);
+		if(normalstdout) DOS_CloseFile(1);
+		if(!normalstdin && !in) DOS_OpenFile("con",2,&dummy);
+		bool status = true;
 		/* Create if not exist. Open if exist. Both in read/write mode */
-		if(!DOS_OpenFileExtended(out,2,2,0x11,&dummy,&dummy2))
-			DOS_OpenFile("con",2,&dummy); //Read only file, open con again
+		if(append) {
+			if( (status = DOS_OpenFile(out,2,&dummy)) ) {
+				 DOS_SeekFile(1,&bigdummy,DOS_SEEK_END);
+			} else {
+				status = DOS_CreateFile(out,2,&dummy); //Create if not exists.
+			}
+		}
+		else 
+			status = DOS_OpenFileExtended(out,2,2,0x12,&dummy,&dummy2);
+		
+		if(!status && normalstdout) DOS_OpenFile("con",2,&dummy); //Read only file, open con again
+		if(!normalstdin && !in) DOS_CloseFile(0);
 	}
 	/* Run the actual command */
 	DoCommand(line);
 	/* Restore handles */
 	if(in) {
 		DOS_CloseFile(0);
-		DOS_OpenFile("con",2,&dummy);
+		if(normalstdin) DOS_OpenFile("con",2,&dummy);
 		free(in);
 	}
-	if(out) {   
+	if(out) {
 		DOS_CloseFile(1);
-		DOS_OpenFile("con",2,&dummy);
+		if(!normalstdin) DOS_OpenFile("con",2,&dummy);
+		if(normalstdout) DOS_OpenFile("con",2,&dummy);
+		if(!normalstdin) DOS_CloseFile(0);
 		free(out);
 	}
 }
@@ -166,7 +268,7 @@ void DOS_Shell::RunInternal(void)
 	while(bf && bf->ReadLine(input_line)) 
 	{
 		if (echo) {
-				if (input_line[0]!='@') {
+				if (input_line[0] != '@') {
 					ShowPrompt();
 					WriteOut(input_line);
 					WriteOut("\n");
@@ -190,7 +292,12 @@ void DOS_Shell::Run(void) {
 		return;
 	}
 	/* Start a normal shell and check for a first command init */
-	WriteOut(MSG_Get("SHELL_STARTUP"));
+	WriteOut(MSG_Get("SHELL_STARTUP_BEGIN"));
+#if C_DEBUG
+	WriteOut(MSG_Get("SHELL_STARTUP_DEBUG"));
+#endif
+	if(machine == MCH_CGA) WriteOut(MSG_Get("SHELL_STARTUP_CGA"));
+	WriteOut(MSG_Get("SHELL_STARTUP_END"));
 	if (cmd->FindString("/INIT",line,true)) {
 		strcpy(input_line,line.c_str());
 		line.erase();
@@ -222,55 +329,91 @@ void DOS_Shell::SyntaxError(void) {
 	WriteOut(MSG_Get("SHELL_SYNTAXERROR"));
 }
 
+class AUTOEXEC:public Module_base {
+private:
+	AutoexecObject autoexec[16];
+	AutoexecObject autoexec_echo;
+public:
+	AUTOEXEC(Section* configuration):Module_base(configuration) {
+		/* Register a virtual AUOEXEC.BAT file */
+		std::string line;
+		Section_line * section=static_cast<Section_line *>(configuration);
 
+		/* add stuff from the configfile unless -noautexec is specified. */
+		char * extra=const_cast<char*>(section->data.c_str());
+		if (extra && !control->cmdline->FindExist("-noautoexec",true)) {
+			/* detect if "echo off" is the first line */
+			bool echo_off  = !strncasecmp(extra,"echo off",8);
+			if (!echo_off) echo_off = !strncasecmp(extra,"@echo off",9);
+
+			/* if "echo off" add it to the front of autoexec.bat */
+			if(echo_off) autoexec_echo.InstallBefore("@echo off");
+
+			/* Install the stuff from the configfile */
+			autoexec[0].Install("%s",extra);
+		}
+
+		/* Check to see for extra command line options to be added (before the command specified on commandline) */
+		/* Maximum of extra commands: 10 */
+		Bitu i = 1;
+		while (control->cmdline->FindString("-c",line,true) && (i <= 11))
+			autoexec[i++].Install((char *)line.c_str());
+	
+		/* Check for the -exit switch which causes dosbox to when the command on the commandline has finished */
+		bool addexit = control->cmdline->FindExist("-exit",true);
+
+		/* Check for first command being a directory or file */
+		char buffer[CROSS_LEN];
+		char cross_filesplit[2] = {CROSS_FILESPLIT , 0};
+		if (control->cmdline->FindCommand(1,line)) {
+			struct stat test;
+			strcpy(buffer,line.c_str());
+			if (stat(buffer,&test)){
+				getcwd(buffer,CROSS_LEN);
+				strcat(buffer,cross_filesplit);
+				strcat(buffer,line.c_str());
+				if (stat(buffer,&test)) goto nomount;
+			}
+			if (test.st_mode & S_IFDIR) { 
+				autoexec[12].Install("MOUNT C \"%s\"",buffer);
+				autoexec[13].Install("C:");
+			} else {
+				char* name = strrchr(buffer,CROSS_FILESPLIT);
+				if (!name) { //Only a filename 
+					line = buffer;
+					getcwd(buffer,CROSS_LEN);
+					strcat(buffer,cross_filesplit);
+					strcat(buffer,line.c_str());
+					if(stat(buffer,&test)) goto nomount;
+					name = strrchr(buffer,CROSS_FILESPLIT);
+					if(!name) goto nomount;
+				}
+				*name++ = 0;
+				if (access(buffer,F_OK)) goto nomount;
+				autoexec[12].Install("MOUNT C \"%s\"",buffer);
+				autoexec[13].Install("C:");
+				upcase(name);
+				if(strstr(name,".BAT") == 0) {
+					autoexec[14].Install(name);
+				} else {
+				/* BATch files are called else exit will not work */
+					char call[CROSS_LEN] = { 0 };
+					strcpy(call,"CALL ");
+					strcat(call,name);
+					autoexec[14].Install(call);
+				}
+				if(addexit) autoexec[15].Install("exit");
+			}
+		}
+nomount:
+		VFILE_Register("AUTOEXEC.BAT",(Bit8u *)autoexec_data,strlen(autoexec_data));
+	}
+};
+
+static AUTOEXEC* test;
 
 void AUTOEXEC_Init(Section * sec) {
-	/* Register a virtual AUOEXEC.BAT file */
-	std::string line;
-	Section_line * section=static_cast<Section_line *>(sec);
-	char * extra=(char *)section->data.c_str();
-	if (extra) SHELL_AddAutoexec("%s",extra);
-	/* Check to see for extra command line options to be added (before the command specified on commandline) */
-	while (control->cmdline->FindString("-c",line,true))
-		SHELL_AddAutoexec((char *)line.c_str());
-	
-	/* Check for the -exit switch which causes dosbox to when the command on the commandline has finished */
-	bool addexit = control->cmdline->FindExist("-exit",true);
-
-	/* Check for first command being a directory or file */
-	char buffer[CROSS_LEN];
-	if (control->cmdline->FindCommand(1,line)) {
-		struct stat test;
-		strcpy(buffer,line.c_str());
-		if (stat(buffer,&test)){
-			getcwd(buffer,CROSS_LEN);
-			strcat(buffer,line.c_str());
-			if (stat(buffer,&test)) goto nomount;
-		}
-		if (test.st_mode & S_IFDIR) { 
-			SHELL_AddAutoexec("MOUNT C \"%s\"",buffer);
-			SHELL_AddAutoexec("C:");
-		} else {
-			char * name=strrchr(buffer,CROSS_FILESPLIT);
-			if (!name) goto nomount;
-			*name++=0;
-			if (access(buffer,F_OK)) goto nomount;
-			SHELL_AddAutoexec("MOUNT C \"%s\"",buffer);
-			SHELL_AddAutoexec("C:");
-			upcase(name);
-			if(strstr(name,".BAT")==0) {
-				SHELL_AddAutoexec(name);
-			} else {
-				char call[CROSS_LEN] = { 0 };
-				strcpy(call,"CALL ");
-				strcat(call,name);
-				SHELL_AddAutoexec(call);
-			}
-			if(addexit) SHELL_AddAutoexec("exit");
-		}
-	}
-nomount:
-	VFILE_Register("AUTOEXEC.BAT",(Bit8u *)autoexec_data,strlen(autoexec_data));
+	test = new AUTOEXEC(sec);
 }
 
 static char * path_string="PATH=Z:\\";
@@ -286,6 +429,7 @@ void SHELL_Init() {
 	MSG_Add("SHELL_CMD_ECHO_OFF","ECHO is off.\n");
 	MSG_Add("SHELL_ILLEGAL_SWITCH","Illegal switch: %s.\n");
 	MSG_Add("SHELL_CMD_CHDIR_ERROR","Unable to change to: %s.\n");
+	MSG_Add("SHELL_CMD_CHDIR_HINT","To change to different drive type \033[31m%c:\033[0m\n");
 	MSG_Add("SHELL_CMD_MKDIR_ERROR","Unable to make: %s.\n");
 	MSG_Add("SHELL_CMD_RMDIR_ERROR","Unable to remove: %s.\n");
 	MSG_Add("SHELL_CMD_DEL_ERROR","Unable to delete: %s.\n");
@@ -302,7 +446,7 @@ void SHELL_Init() {
 	MSG_Add("SHELL_CMD_DIR_INTRO","Directory of %s.\n");
 	MSG_Add("SHELL_CMD_DIR_BYTES_USED","%5d File(s) %17s Bytes.\n");
 	MSG_Add("SHELL_CMD_DIR_BYTES_FREE","%5d Dir(s)  %17s Bytes free.\n");
-	MSG_Add("SHELL_EXECUTE_DRIVE_NOT_FOUND","Drive %c does not exist!\nYou must \033[31mmount\033[0m it first. Type \033[1;33mintro\033[0m for more information.\n");
+	MSG_Add("SHELL_EXECUTE_DRIVE_NOT_FOUND","Drive %c does not exist!\nYou must \033[31mmount\033[0m it first. Type \033[1;33mintro\033[0m or \033[1;33mintro mount\033[0m for more information.\n");
 	MSG_Add("SHELL_EXECUTE_ILLEGAL_COMMAND","Illegal command: %s.\n");
 	MSG_Add("SHELL_CMD_PAUSE","Press any key to continue.\n");
 	MSG_Add("SHELL_CMD_PAUSE_HELP","Waits for 1 keystroke to continue.\n");
@@ -311,7 +455,7 @@ void SHELL_Init() {
 	MSG_Add("SHELL_CMD_SUBST_NO_REMOVE","Removing drive not supported. Doing nothing.\n");
 	MSG_Add("SHELL_CMD_SUBST_FAILURE","SUBST failed. You either made an error in your commandline or the target drive is already used.\nIt's only possible to use SUBST on Local drives");
 
-	MSG_Add("SHELL_STARTUP",
+	MSG_Add("SHELL_STARTUP_BEGIN",
 		"\033[44;1m\xC9\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD"
 		"\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD"
 		"\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xBB\n"
@@ -324,18 +468,23 @@ void SHELL_Init() {
 		"\xBA To activate the keymapper \033[31mctrl-F1\033[37m.                                 \xBA\n"
 		"\xBA For more information read the \033[36mREADME\033[37m file in the DOSBox directory. \xBA\n"
 		"\xBA                                                                    \xBA\n"
-#if C_DEBUG
-		"\xBA Press \033[31mPause\033[37m to enter the debugger or start the exe with \033[33mDEBUG\033[37m.     \xBA\n"
-		"\xBA                                                                    \xBA\n"
-#endif
-		"\xBA \033[32mHAVE FUN!\033[37m                                                          \xBA\n"
-		"\xBA \033[32mThe DOSBox Team\033[37m                                                    \xBA\n"
-		"\xC8\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD"
-		"\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD"
-		"\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xBC\033[0m\n"
-		//"\n" //Breaks the startup message if you type a mount and a drive change.
 	);
-
+	MSG_Add("SHELL_STARTUP_CGA","\xBA DOSBox supports Composite CGA mode.                                \xBA\n"
+	        "\xBA Use \033[31m(alt-)F11\033[37m to change the colours when in this mode.             \xBA\n"
+	        "\xBA                                                                    \xBA\n"
+	);
+	MSG_Add("SHELL_STARTUP_DEBUG",
+	        "\xBA Press \033[31malt-Pause\033[37m to enter the debugger or start the exe with \033[33mDEBUG\033[37m. \xBA\n"
+	        "\xBA                                                                    \xBA\n"
+	);
+	MSG_Add("SHELL_STARTUP_END",
+	        "\xBA \033[32mHAVE FUN!\033[37m                                                          \xBA\n"
+	        "\xBA \033[32mThe DOSBox Team\033[37m                                                    \xBA\n"
+	        "\xC8\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD"
+	        "\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD"
+	        "\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xCD\xBC\033[0m\n"
+	        //"\n" //Breaks the startup message if you type a mount and a drive change.
+	);
 	MSG_Add("SHELL_CMD_CHDIR_HELP","Change Directory.\n");
 	MSG_Add("SHELL_CMD_CLS_HELP","Clear screen.\n");
 	MSG_Add("SHELL_CMD_DIR_HELP","Directory View.\n");
@@ -371,11 +520,20 @@ void SHELL_Init() {
 	PROGRAMS_MakeFile("COMMAND.COM",SHELL_ProgramStart);
 
 	/* Now call up the shell for the first time */
-	Bit16u psp_seg=DOS_GetMemory(16+1)+1;
+	Bit16u psp_seg=DOS_GetMemory(16+3)+1;
 	Bit16u env_seg=DOS_GetMemory(1+(4096/16))+1;
 	Bit16u stack_seg=DOS_GetMemory(2048/16);
 	SegSet16(ss,stack_seg);
 	reg_sp=2046;
+
+	/* Set up int 24 and psp (Telarium games) */
+	real_writeb(psp_seg+16+1,0,0xea);		/* far jmp */
+	real_writed(psp_seg+16+1,1,real_readd(0,0x24*4));
+	real_writed(0,0x24*4,((Bit32u)psp_seg<<16) | ((16+1)<<4));
+
+	/* Set up int 23 to "int 20" in the psp. Fixes what.exe */
+	real_writed(0,0x23*4,((Bit32u)psp_seg<<16));
+
 	/* Setup MCB and the environment */
 	DOS_MCB envmcb((Bit16u)(env_seg-1));
 	envmcb.SetPSPSeg(psp_seg);
@@ -394,12 +552,20 @@ void SHELL_Init() {
 	DOS_PSP psp(psp_seg);
 	psp.MakeNew(0);
 	dos.psp(psp_seg);
+   
+	/* The start of the filetable in the psp must look like this:
+	 * 01 01 01 00 02
+	 * In order to achieve this: First open 2 files. Close the first and
+	 * duplicate the second (so the entries get 01) */
 	Bit16u dummy=0;
 	DOS_OpenFile("CON",2,&dummy);/* STDIN  */
 	DOS_OpenFile("CON",2,&dummy);/* STDOUT */
-	DOS_OpenFile("CON",2,&dummy);/* STDERR */
+	DOS_CloseFile(0);            /* Close STDIN */
+	DOS_ForceDuplicateEntry(1,0);/* "new" STDIN */
+	DOS_ForceDuplicateEntry(1,2);/* STDERR */
 	DOS_OpenFile("CON",2,&dummy);/* STDAUX */
 	DOS_OpenFile("CON",2,&dummy);/* STDPRN */
+
 	psp.SetParent(psp_seg);
 	/* Set the environment */
 	psp.SetEnvironment(env_seg);
@@ -413,8 +579,9 @@ void SHELL_Init() {
 	dos.dta(RealMake(psp_seg,0x80));
 	dos.psp(psp_seg);
 
-	Program * new_program;
-	SHELL_ProgramStart(&new_program);
-	new_program->Run();
-	delete new_program;
+	
+	SHELL_ProgramStart(&first_shell);
+	first_shell->Run();
+	delete first_shell;
+	first_shell = 0;//Make clear that it shouldn't be used anymore
 }
